#include "journalfile.hpp"

namespace gjournal
{
	JournalFile::JournalFile( JournalWindow *window )
	{
		srand( time( 0 ) );

		this->window = window;
	}

	bool JournalFile::_createNewFile( std::string file, std::string &error )
	{
		sqlitemm::SQliteManager db;

		try {
			db.connect( file );
		}
		catch( sqlitemm::SQliteConnectionException ex )
		{
			error = ex.what();
			return false;
		}

		std::string queries[3];
		queries[0] = "CREATE TABLE entries( Date DATE, Entry MEDIUMTEXT, PRIMARY KEY( Date ) );";
		queries[1] = "CREATE TABLE version( revision VARCHAR(10) );";
		queries[2] = "INSERT INTO version VALUES( '0.3.5' );";

		try {
			for( int i = 0; i < 3; i++ )
			{
				db.query( queries[i] );
			}
		}
		catch( sqlitemm::SQliteQueryException ex )
		{
			error = ex.what();
			return false;
		}

		db.disconnect();
		#ifndef WIN32
			chmod( file.c_str(), 0600 );
		#endif


		return true;
	}

	bool JournalFile::isJournalFile( std::string file )
	{
		char temp[1024] = { '\0' };
		unsigned char checksig[25] = { '\0' };

		std::ifstream in( file.c_str(), std::ios::binary );
		if( !in.is_open() ) return false;

		// get the various required bytes from the file for decryption
		in.getline( temp, 1024 );
		in.getline( temp, 5 );
		in.read( (char *)checksig, 25 );
		int size = in.gcount();
		if( !size ) return false;
		in.close();

		// check to make sure the magic numbers were present
		if( strcmp( (char *)aolib::BFCryptFile::encsig, (char *)checksig ) != 0 )
		{
			return false;
		}

		return true;
	}

	void JournalFile::_loadFileHandler( const char *f )
	{
		std::string oldTempFile = window->tempFile;
		std::string file( f );

		if( !JournalFile::isJournalFile( file ) )
		{
			Gtk::MessageDialog dialog( "The file you specified is not a Gtk Journal file, or it has been corrupted.", Gtk::MESSAGE_WARNING );
			dialog.run();
			dialog.hide();
			if( window->loading )
			{
				window->loading = false;
				window->_newFileHandler();
			}

			return;
		}

		aolib::LoginDialog login( "Please enter the username and password" );
		login.set_icon( Gdk::Pixbuf::create_from_xpm_data( gtk_journal_icon_xpm ) );
		bool check = false;

		if( !Glib::file_test( file, Glib::FILE_TEST_EXISTS ) )
		{
			Gtk::MessageDialog dialog( "Sorry, the file you specified does not exist.", Gtk::MESSAGE_WARNING );
			dialog.run();
			dialog.hide();
			if( window->loading )
			{
				window->loading = false;
				window->_newFileHandler();
			}

			return;
		}

		if( window->sql.isConnected() ) window->sql.disconnect();

		int result;
		window->tempFile = _getNewTempFile();

		while( !check )
		{
			result = login.run();
			if( result != Gtk::RESPONSE_OK )
			{
				if( window->loading )
				{
					window->_shredTempFile( oldTempFile );
					exit( -1 );
				}

				check = true;
				window->tempFile = oldTempFile;
			}
			else {
				Gtk::MessageDialog pd( "Decrypting Journal, please wait...", Gtk::MESSAGE_INFO, Gtk::BUTTONS_NONE );
				pd.set_modal();
				pd.show();

				while( Gtk::Main::events_pending() )
				{
					Gtk::Main::iteration();
				}


				try {
					aolib::BFCryptFile::decryptFile( login.getUsername(), login.getPassword(), file, window->tempFile );
					check = true;
				}
				catch( aolib::BFCryptException ex )
				{
					Gtk::MessageDialog dialog( ex.what(), Gtk::MESSAGE_WARNING );
					dialog.run();
					dialog.hide();
					check = false;
				}
				pd.hide();
			}
		}

		if( result != Gtk::RESPONSE_OK )
		{
			window->tempFile = oldTempFile;
			window->_dbConnect();
			return;
		}

		Globals::dbSettings["password"] = login.getPassword();
		Globals::dbSettings["username"] = login.getUsername();

		window->currentEntry = "";
		window->loading = false;
		window->journalFile = file;
		window->_shredTempFile( oldTempFile );
		window->_dbConnect();

		// update the SQL database to the newest version
		VersionController::checkVersion( &window->sql );

		window->_todayHandler();
		window->_markMonthDays();
		window->savedFlag = true;
		std::string title = "Gtk+ Journal - ";
		std::string name = window->journalFile.c_str();
		unsigned int b = name.find_last_of( Globals::separator );
		if( b != std::string::npos )
		{
			name = name.substr( b + 1, name.length() );
		}

		if( window->journalFile != "" )	title += name;
		else title += "New Journal";
		window->set_title( title );
		window->statusBar.pop();
		window->statusBar.push( "Status: Unmodified" );
	}

	std::string JournalFile::_getNewTempFile()
	{
		std::stringstream tmpStream;
		std::string tempFile;

		tmpStream << ".tmp___gtkj___" << rand() << rand();
		tempFile = tmpStream.str();
		tempFile = Globals::SETTINGS_DIR + Globals::separator + tempFile;

		return tempFile;
	}

	void JournalFile::_preferences( bool saving )
	{
		if( saving && Globals::dbSettings["password"] == "" )
		{
			Gtk::MessageDialog dialog( *window, "Since this is a new Journal, you must set your username and password before you can save.",
				Gtk::MESSAGE_INFO );
			dialog.run();
			dialog.hide();
		}

		// display the informationi dialog and populate it with
		// the information from the SQL query
		PreferencesDialog information( window, &window->sql );
		information.show();
		int res = information.run();

		if( res == Gtk::RESPONSE_OK )
		{
			if( saving ) window->_saveHandler();
			else {
				window->_journalEntryChangeHandler();
			}
		}
	}
}
