/*
    Copyright (C) 2004 Adam Olsen

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <gjournal/journalwindow.hpp>

GError *gtkSpellError;

namespace gjournal
{
	JournalWindow::JournalWindow( const char *loadFile )
	: dateLabel( "Entry Date: " ), savedFlag( true )
	{
		jf = new JournalFile( this );
		_checkForTemps();
		_checkForOldFileFormat();

		// set the icon for the window
		set_icon( Gdk::Pixbuf::create_from_xpm_data( gtk_journal_icon_xpm ) );

		// layout the child components
		_layoutComponents();

		// set up the various signal handlers
		signal_delete_event().connect( SigC::slot( *this, &JournalWindow::_closeHandler ) );
		calendar.signal_day_selected().connect( SigC::slot( *this, &JournalWindow::_daySelectHandler ) );
		modified = journalEntry.get_buffer()->signal_changed().connect( SigC::slot( *this, &JournalWindow::_journalEntryChangeHandler ) );
		calendar.signal_next_month().connect( SigC::slot( *this, &JournalWindow::_markMonthDays ) );
		calendar.signal_prev_month().connect( SigC::slot( *this, &JournalWindow::_markMonthDays ) );
		calendar.signal_next_year().connect( SigC::slot( *this, &JournalWindow::_markMonthDays ) );
		calendar.signal_prev_year().connect( SigC::slot( *this, &JournalWindow::_markMonthDays ) );

		if( !loadFile )
		{
			_newFileHandler();
			loading = false;
		}
		else {
			loading = true;
			jf->_loadFileHandler( loadFile );
		}

		#ifdef HAVE_GTKSPELL
			gtkSpell = 0;
		#endif

		show_all_children();
		show();
		PreferencesDialog::loadSettings( this );
	}

	sqlitemm::SQliteManager *JournalWindow::getDB()
	{
		if( !sql.isConnected() ) return 0;
		return &sql;
	}

#ifdef HAVE_GTKSPELL
	bool JournalWindow::usingGtkSpell()
	{
		if( gtkSpell ) return true;
		else return false;
	}

	void JournalWindow::detachGtkSpell()
	{
		gtkspell_detach( gtkSpell );
		gtkSpell = 0;
	}

	void JournalWindow::attachGtkSpell()
	{
		gtkSpell = gtkspell_new_attach( journalEntry.gobj(), NULL, &gtkSpellError );

		if( !gtkSpell )
		{
			g_print( gtkSpellError->message );
			g_print( "\n" );
			g_error_free( gtkSpellError );
		}
	}
#endif


	/**
	 * Checks to see if the user is running the 0.3.2 file format (the one that didn't support
	 * multiple files), and tells them that they need to conver the file
	**/
	void JournalWindow::_checkForOldFileFormat()
	{
		// journal.enc in ~/.gtkjournal will signify an old file format
		std::string oldJournal = Globals::SETTINGS_DIR + Globals::separator + "journal.enc";

		if( Glib::file_test( oldJournal, Glib::FILE_TEST_EXISTS ) )
		{
			Glib::ustring mess = "It looks like you haven't yet converted your journal file from the old format yet. ";
			mess += "You will need to do this before you can access your old journal.  ";
			mess += "You can find information on how in the sourcecode's README file.  Please remove your \"~/.gtkjournal\" directory ";
			mess += "if you have already converted your old journal file and you don't want to see this message again.";
			Gtk::MessageDialog dialog( mess, Gtk::MESSAGE_INFO );
			dialog.run();
			dialog.hide();
		}
	}

	/**
	 * Finds all the days that have entries for this month, and then marks them on the calendar
	**/
	void JournalWindow::_markMonthDays()
	{
		std::string query;
		std::stringstream yd;
		guint month, day, year;

		// get the current date
		calendar.get_date( year, month, day );
		calendar.clear_marks();

		// format the date for searching the database
		yd << year << "-" << std::setw( 2 ) << std::setfill( '0' ) << ( month + 1 );

		// build the query
		query = "SELECT Date FROM entries WHERE Date <= '" + yd.str() + "-31' AND Date >= '" + yd.str() + "-00'";

		sqlitemm::SQliteResult res;
		try {
			res = sql.query( query );
		}
		catch( sqlitemm::SQliteQueryException ex )
		{
			// here it's ok if the query is bad, but print an error to STDOUT
			g_print( ex.what().c_str() );
			return;
		}

		// if there's no results for this month, we don't need to mark any days
		if( !res.numRows() ) return;

		while( res.hasNext() )
		{
			std::map<std::string, std::string> row = res.next();

			yd.clear();
			yd.str( row["Date"].substr( 8, 2 ) );
			yd >> day; // make the string an int

			// mark the day
			calendar.mark_day( day );
		}
	}

	/**
	 * Creates a new empty temp journal file
	**/
	void JournalWindow::_newFileHandler()
	{
		if( !savedFlag )
		{
			// if they haven't saved their current journal, ask them if they want
			// to do so
			Gtk::MessageDialog dialog( "Save your current Journal?", Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
			int result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		// if a temp file exists, write 0s to it
		if( tempFile != "" ) _shredTempFile( tempFile );


		if( sql.isConnected() ) sql.disconnect();

		// get a new tempfile
		tempFile = jf->_getNewTempFile();

		std::string error;

		// actually create the new file with the correct SQL
		if( !jf->_createNewFile( tempFile, error ) )
		{
			Gtk::MessageDialog dialog( "Could not create temp file.\n" + error, Gtk::MESSAGE_ERROR );
			dialog.run();
			dialog.hide();
			if( Globals::dbSettings["password"] == "" ) exit( -1 );
			return;
		}
		else {

			// set the current text buffer to nothing
			Glib::RefPtr<Gtk::TextBuffer> buf = journalEntry.get_buffer();
			buf->set_text( "" );
			journalEntry.set_buffer( buf );
			journalFile = "";

			// connect to the database, and jump to today on the calendar
			_dbConnect();
			_todayHandler();
			_markMonthDays();
		}

		// delete the global username and password
		Globals::dbSettings["username"] = "";
		Globals::dbSettings["password"] = "";

		statusBar.pop();
		statusBar.push( "Status: Unmodified" );
		std::string title = "Gtk+ Journal - *New Journal";
		set_title( title );
		savedFlag = true;
	}

	/**
	 * Selects the current day, and loads the days entry from the
	 * database
	**/
	void JournalWindow::_daySelectHandler()
	{
		guint month, day, year;

		// get the day that is currently selected
		calendar.get_date( year, month, day );

		// and load the entry for it
		_loadEntry( ( month + 1 ), day, year );
		_saveCurrentEntry();
	}

	void JournalWindow::_openHandler()
	{
		if( !savedFlag )
		{
			// if the file has not yet been saved yet, find out if they want to save it
			Gtk::MessageDialog dialog( "Save your current Journal?", Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
			int result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				// save the file
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		// show a file selection box
		Gtk::FileSelection fs( "Open Journal File" );
		fs.set_transient_for( *this );
		int result = fs.run();
		if( result == Gtk::RESPONSE_OK )
		{
			fs.hide();

			// if they clicked ok, attempt to load the journal file
			jf->_loadFileHandler( fs.get_filename().c_str() );
		}
	}

	void JournalWindow::_layoutComponents()
	{
		add( mainBox );
		mainBox.pack_start( menuBar, Gtk::PACK_SHRINK );

		// File Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::MenuElem( "_File", fileMenu ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::NEW, SigC::slot( *this, &JournalWindow::_newFileHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::OPEN, SigC::slot( *this, &JournalWindow::_openHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::SAVE, SigC::slot( *this, &JournalWindow::_saveHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::SAVE_AS, SigC::slot( *this, &JournalWindow::_saveAsHandler ) ) );

		fileMenu.items().push_back( Gtk::Menu_Helpers::SeparatorElem() );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::QUIT, SigC::slot( *this, &JournalWindow::_exitHandler ) ) );

		// Edit Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::MenuElem( "_Edit", editMenu ) );
		editMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::PREFERENCES, SigC::slot( *this, &JournalWindow::_preferencesHandler ) ) );

		// Help Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::HELP, helpMenu ) );
		helpMenu.items().push_back( Gtk::Menu_Helpers::MenuElem( "_About", SigC::slot( *this, &JournalWindow::_aboutHandler ) ) );

		// create a toolbar with three items
		mainBox.pack_start( toolBar, Gtk::PACK_SHRINK );
		{
			using namespace Gtk::Toolbar_Helpers;

			// create some buttons
			StockElem newItem( Gtk::Stock::NEW, SigC::slot( *this, &JournalWindow::_newFileHandler ), "New Journal" );
			StockElem open( Gtk::Stock::OPEN, SigC::slot( *this, &JournalWindow::_openHandler ), "Open Journal" );
			StockElem save( Gtk::Stock::SAVE, SigC::slot( *this, &JournalWindow::_saveHandler ), "Save Journal" );
			StockElem today( Gtk::Stock::JUMP_TO, SigC::slot( *this, &JournalWindow::_todayHandler ), "Today's Entry" );

			// add the buttons
			toolBar.tools().push_back( newItem );
			toolBar.tools().push_back( open );
			toolBar.tools().push_back( save );
			toolBar.tools().push_back( Space() );
			toolBar.tools().push_back( today );

			// add the date label
			Element date( dateLabel );
			toolBar.tools().push_back( date );
		}

		calendarBox.set_size_request( 210, 70 );

		mainBox.pack_start( centerBox );
		centerBox.pack_start( calendarBox, Gtk::PACK_SHRINK );
		calendarBox.pack_start( calendar, Gtk::PACK_SHRINK );
		window.add( journalEntry );
		journalEntry.set_wrap_mode( Gtk::WRAP_WORD );
		window.set_shadow_type( Gtk::SHADOW_ETCHED_IN );
		window.set_policy( Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC );
		centerBox.pack_start( window );

		mainBox.pack_start( statusBar, Gtk::PACK_SHRINK );

		set_default_size( 750, 480 );
		// center the journal window
		set_position( Gtk::WIN_POS_CENTER );
	}

	void JournalWindow::_loadEntry( int month, int mday, int year )
	{
		std::stringstream date;

		if( currentEntry != "" ) _saveCurrentEntry();

		// format the date in a string like 2004-01-05
		date << year << "-";
		date << std::setw( 2 ) << std::setfill( '0' ) << month << "-";
		date << std::setw( 2 ) << std::setfill( '0' ) << mday;
		std::string query = "SELECT Date, Entry FROM entries WHERE Date=\"" + date.str() + "\";";

		// we need to disconnect the modified signal handler because switching days
		// and clearing out the buffer will set it off
		modified.disconnect();

		// select for today from sqlite

		bool success = true;
		sqlitemm::SQliteResult res;
		try {
			res = sql.query( query );
		}
		catch( sqlitemm::SQliteQueryException ex )
		{
			_sqlError( ex.what() );
			success = false;
		}

		if( success )
		{
			currentEntry = date.str();
			date.clear();
			date.str( "" );
			date << aolib::long_month_from_int( month ) << " " << mday << ", " << year;

			// set the date label to the date being loaded
			dateLabel.set_text( " Entry Date: " + date.str() );

			// put focus into the entry box
			journalEntry.grab_focus();
			Glib::RefPtr<Gtk::TextBuffer> buf = journalEntry.get_buffer();

			if( !res.numRows() ) buf->set_text( "" );
			else {
				std::map<std::string, std::string> row = res.next();
				buf->set_text( row["Entry"] );
			}

			journalEntry.set_buffer( buf );
		}

		// reconnect the modified listener
		modified = journalEntry.get_buffer()->signal_changed().connect( SigC::slot( *this, &JournalWindow::_journalEntryChangeHandler ) );
	}

	void JournalWindow::_journalEntryChangeHandler()
	{
		// mark the journal as not saved
		savedFlag = false;
		std::string title = "Gtk+ Journal - *";

		std::string name = journalFile.c_str();
		unsigned int b = name.find_last_of( Globals::separator );
		if( b != std::string::npos )
		{
			name = name.substr( b + 1, name.length() );
		}

		if( journalFile != "" )	title += name;
		else title += "New Journal";
		set_title( title );
		statusBar.pop();
		statusBar.push( "Status: Modified (not saved)" );
	}

	void JournalWindow::_saveCurrentEntry()
	{
		std::stringstream sday;
		sday << currentEntry.substr( 8, 10 );
		guint day;
		sday >> day;

		try {
			if( journalEntry.get_buffer()->get_text() == "" )
			{
				sql.go( "DELETE FROM entries WHERE Date='" + currentEntry + "'" );
				calendar.unmark_day( day );
				return;
			}

			std::string entry = sqlitemm::SQliteManager::addSlashes( journalEntry.get_buffer()->get_text() );
			std::string query = "REPLACE INTO entries( Date, Entry ) VALUES( \"" + currentEntry + "\", \"" + entry + "\" );";

			// save the current entry into the database
			sql.go( query );
			calendar.mark_day( day );
		}
		catch( sqlitemm::SQliteQueryException ex )
		{
			_sqlError( ex.what() );
		}
	}

	void JournalWindow::_dbConnect()
	{
		try {
			sql.connect( tempFile );
		}
		catch( sqlitemm::SQliteConnectionException ex )
		{
			_sqlError( ex.what(), true );
		}
	}

	void JournalWindow::_saveAsHandler()
	{
		if( Globals::dbSettings["password"] == "" )
		{
			jf->_preferences( true );
			return;
		}

		bool check = false;
		Gtk::FileSelection fs( "Save As" );
		fs.set_transient_for( *this );

		while( !check )
		{
			int result = fs.run();
			if( result != Gtk::RESPONSE_OK ) return;

			if( Glib::file_test( fs.get_filename(), Glib::FILE_TEST_EXISTS ) )
			{
				Gtk::MessageDialog dialog( fs, "The filename you have chosen already exists.  Do you want to overrwite it?", Gtk::MESSAGE_QUESTION,
					Gtk::BUTTONS_YES_NO );
				int res = dialog.run();
				if( res == Gtk::RESPONSE_YES ) check = true;
			}
			else check = true;
		}

		journalFile = fs.get_filename();
		_saveHandler();
	}

	void JournalWindow::_saveHandler()
	{
		if( currentEntry != "" ) _saveCurrentEntry();

		if( journalFile == "" )
		{
			_saveAsHandler();
			return;
		}

		sql.disconnect();
		Gtk::MessageDialog pd( "Encrypting Journal, please wait...", Gtk::MESSAGE_INFO, Gtk::BUTTONS_NONE );
		pd.set_modal();
		pd.show();

		while( Gtk::Main::events_pending() )
		{
			Gtk::Main::iteration();
		}

		bool success = true;
		try {
			aolib::BFCryptFile::encryptFile(  Globals::dbSettings["username"], Globals::dbSettings["password"], tempFile, journalFile );
		}
		catch( aolib::BFCryptException ex )
		{
			Gtk::MessageDialog dialog( *this, ex.what(), Gtk::MESSAGE_ERROR );
			dialog.run();
			success = false;
		}

		if( success )
		{
			savedFlag = true;
			std::string title = "Gtk+ Journal - ";
			std::string name = journalFile.c_str();
			unsigned int b = name.find_last_of( Globals::separator );
			if( b != std::string::npos )
			{
				name = name.substr( b + 1, name.length() );
			}

			if( journalFile != "" )	title += name;
			else title += "New Journal";
			set_title( title );
			statusBar.pop();
			statusBar.push( "Status: Unmodified (saved)" );
		}

		pd.hide();

		// and then reconned to the temp database
		_dbConnect();
	}

	void JournalWindow::_todayHandler()
	{
		// get the current date
		time_t t = time( NULL );
		tm *the = localtime( &t );

		// load the entry from the current date
		_loadEntry( ( the->tm_mon + 1 ), the->tm_mday, ( the->tm_year + 1900 ) );
		// select the day specified in the calendar
		calendar.select_month( the->tm_mon, ( the->tm_year + 1900 ) );
		calendar.select_day( the->tm_mday );
	}

	void JournalWindow::_preferencesHandler()
	{
		jf->_preferences();
	}

	void JournalWindow::_sqlError( std::string error, bool e )
	{
		if( error == "" ) error = "Unknown SQL error";

		// display the error
		Gtk::MessageDialog dialog( *this, "There has been an error executing your request.\n" + error, Gtk::MESSAGE_ERROR );
		dialog.run();

		// if it's a fatal error, bail out
		if( e ) exit( -1 );
	}

	void JournalWindow::_exitHandler()
	{
		int result = Gtk::RESPONSE_NO;

		// if they haven't saved, ask them if they really want to exit
		if( !savedFlag )
		{
			// find out if they want to quit
			Gtk::MessageDialog dialog( *this, "Would you like to save your Journal before quitting?",
				Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );

			result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		sql.disconnect(); // close the database connection
		hide();

		_shredTempFile( tempFile );
		PreferencesDialog::saveSettings();
		g_print( "Goodbye!\n" );
		exit( 0 );
	}

	void JournalWindow::_shredTempFile( std::string file )
	{
		Gtk::MessageDialog dialog( "Shredding temp file, please wait...", Gtk::MESSAGE_INFO, Gtk::BUTTONS_NONE );
		dialog.set_modal();
		dialog.show();

		while( Gtk::Main::events_pending() )
		{
			Gtk::Main::iteration();
		}

		dialog.hide();

		if( Glib::file_test( file, Glib::FILE_TEST_EXISTS ) ) aolib::BFCryptFile::shredFile( file.c_str() );
	}

	void JournalWindow::_checkForTemps()
	{
		if( !Glib::file_test( Globals::SETTINGS_DIR, Glib::FILE_TEST_IS_DIR ) )
		{
			#ifdef WIN32
				mkdir( Globals::SETTINGS_DIR.c_str() );
			#else
				mkdir( Globals::SETTINGS_DIR.c_str(), 0700 );
			#endif
		}

		DIR *dir = opendir( Globals::SETTINGS_DIR.c_str() );
		dirent *d;
		bool warned = false;
		while( ( d = readdir( dir ) ) != 0 )
		{
			std::string filename = d->d_name;
			unsigned int b = filename.find( ".tmp___gtkj___", 0 );
			if( b != std::string::npos )
			{
				if( !warned )
				{
					Glib::ustring message = "It appears as though Gtk+ Journal was exited abnormally.  You should always try to exit ";
					message += "correctly by choosing 'File->Quit' or by typing CTRL+Q, because not doing so can leave your data unprotected.";

					Gtk::MessageDialog dialog( *this, message );
					dialog.run();
					dialog.hide();
					warned = true;
				}

				_shredTempFile( Globals::SETTINGS_DIR + Globals::separator + filename );
			}
		}
		closedir( dir );
	}

	bool JournalWindow::_closeHandler( GdkEventAny *event )
	{
		// basically just call the exit handler
		_exitHandler();
		return true;
	}

	void JournalWindow::_aboutHandler()
	{
		std::string ver( VERSION );
		Gtk::MessageDialog about( *this, "Gtk+ Journal " + ver + "\nBy: Adam R Olsen (adamolsen@net-bizz.com)\n(c)2004" );
		about.run();
	}

	JournalWindow::~JournalWindow()
	{
		delete jf;
	}
}
