/*
    Copyright (C) 2004 Adam Olsen

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <gjournal/preferencesdialog.hpp>

namespace gjournal
{
	PreferencesDialog::PreferencesDialog( JournalWindow *window, sqlitemm::SQliteManager *sql )
	: Gtk::Dialog( "Settings" ), applyButton( Gtk::Stock::APPLY ), okButton( Gtk::Stock::OK )
	{
		set_icon( Gdk::Pixbuf::create_from_xpm_data( gtk_journal_icon_xpm ) );
		this->sql = sql;
		this->window = window;

		// lay out the components so they look nice
		_layoutComponents();

		add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );

		get_action_area()->pack_start( applyButton );
		get_action_area()->pack_start( okButton );
		okButton.signal_clicked().connect( SigC::slot( *this, &PreferencesDialog::_okHandler ) );
		signal_response().connect( SigC::slot( *this, &PreferencesDialog::_responseHandler ) );
		applyButton.signal_clicked().connect( SigC::slot( *this, &PreferencesDialog::_applyHandler ) );

		// center the window
		set_position( Gtk::WIN_POS_CENTER );

		// show the windows children
		show_all_children();
		set_resizable( false );
	}

	void PreferencesDialog::_okHandler()
	{
		for( int i = 0; i < notebook.get_n_pages(); i++ )
		{
			Gtk::Widget *widget = notebook.get_nth_page( i );
			PrefsTabInterface *tab = dynamic_cast<PrefsTabInterface *>( widget );
			if( !tab->_save() ) return;
		}

		response( Gtk::RESPONSE_OK );
	}

	void PreferencesDialog::_applyHandler()
	{
		int i = notebook.get_current_page();
		Gtk::Widget *widget = notebook.get_nth_page( i );
		PrefsTabInterface *tab = dynamic_cast<PrefsTabInterface *>( widget );
		tab->_save();
	}

	void PreferencesDialog::_responseHandler( int response )
	{
		hide();
	}

	void PreferencesDialog::_layoutComponents()
	{
		// set up all the components so they look nice
		get_vbox()->pack_start( notebook );

		passwordTab = new PasswordPrefsTab( this );
		generalTab = new GeneralPrefsTab( this );

		// add the first tab
		notebook.prepend_page( *generalTab, "General Settings" );
		notebook.prepend_page( *passwordTab, "Username and Password" );
	}

	Gtk::TextView *PreferencesDialog::getJournalEntry()
	{
		return window->getJournalEntry();
	}

	JournalWindow *PreferencesDialog::getWindow()
	{
		return window;
	}

	PreferencesDialog::~PreferencesDialog()
	{
		delete passwordTab;
		delete generalTab;
	}

	/**
	 * Static method to save the settings file information
	 **/
	void PreferencesDialog::saveSettings()
	{
		std::ofstream out( Globals::SETTINGS_FILE.c_str() );
		if( !out.is_open() )
		{
			g_print( "Could not write settings file.\n" );
			return;
		}

		std::map<std::string, std::string>::iterator i;
		for( i = Globals::settings.begin(); i != Globals::settings.end(); i++ )
		{
			out << i->first << "=" << Globals::settings[i->first] << std::endl;
		}

		out.close();
	}

	/**
	 * Static method to load the settings file information into Globals::dbSettings
	**/
	void PreferencesDialog::loadSettings( JournalWindow *window )
	{
		char buf[2049] = { '\0' };
		std::ifstream in( Globals::SETTINGS_FILE.c_str() );

		if( !in.is_open() ) return;

		Globals::settings.clear();

		while( !in.eof() )
		{
			in.getline( buf, 2049 );
			if( !in.gcount() ) break;

			std::string line( buf );

			unsigned int b = line.find( "=" );
			if( b != std::string::npos )
			{
				Globals::settings[line.substr( 0, b )] = line.substr( b + 1, line.length() );
			}
		}

		in.close();

		if( Globals::settings["font"] != "" )
		{
			Pango::FontDescription font( Globals::settings["font"] );
			window->getJournalEntry()->modify_font( font );
		}

		#ifdef HAVE_GTKSPELL
			if( Globals::settings["useGtkSpell"] == "true" )
			{
				if( !window->usingGtkSpell() ) window->attachGtkSpell();
			}
			else {
				if( window->usingGtkSpell() ) window->detachGtkSpell();
			}
		#endif
	}

	/**
	 * Load the database settings
	**/
	void PreferencesDialog::loadDBSettings( JournalWindow *window )
	{
		std::string username = Globals::dbSettings["username"];
		std::string password = Globals::dbSettings["password"];

		Globals::dbSettings.clear();

		sqlitemm::SQliteManager *sql = window->getDB();
		if( !sql ) return;
		sqlitemm::SQliteResult res;

		try {
			res = sql->query( "SELECT * FROM settings" );
			std::map<std::string, std::string> row = res.next();
			std::map<std::string, std::string>::iterator i;

			for( i = row.begin(); i != row.end(); i++ )
			{
				Globals::dbSettings[i->first] = row[i->first];
			}
		}
		catch( sqlitemm::SQliteException ex )
		{
			Gtk::MessageDialog dialog( *window, "There was an error loading your settings" + ex.what(), Gtk::MESSAGE_ERROR );
			dialog.run();
		}

		Globals::settings["username"] = username;
		Globals::settings["password"] = password;
	}
}
